#include <udm_config.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_ZLIB
#include <zlib.h>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_env.h"
#include "udm_agent.h"
#include "udm_conf.h"
#include "udm_log.h"
#include "udm_services.h"
#include "udm_store.h"
#include "udm_xmalloc.h"

#define UDM_STORED_MAXCLNTS 128

static char pidname[1024];
static char time_pid[100];
static UDM_ENV * Conf = NULL;
static UDM_AGENT * Agent = NULL;
static int verb = -1;


static void exitproc(void){
	unlink(pidname);
}

static char * time_pid_info(void){
	struct tm * tim;
	time_t t;
	t=time(NULL);
	tim=localtime(&t);
	strftime(time_pid,sizeof(time_pid),"%a %d %T",tim);
	sprintf(time_pid+strlen(time_pid)," [%d]",(int)getpid());
	return(time_pid);
}

/********************* SIG Handlers ****************/
static void sighandler(int sign);
static void init_signals(void){
	/* Set up signals handler*/
	signal(SIGPIPE,sighandler);
	signal(SIGHUP,sighandler);
	signal(SIGINT,sighandler);
	signal(SIGTERM,sighandler);
	signal(SIGCHLD, sighandler);
}
static int have_sighup=0;
static int have_sigint=0;
static int have_sigterm=0;
static int have_sigpipe=0;

static void sighandler(int sign){
#ifdef UNIONWAIT
	union wait status;
#else
	int status;
#endif

	switch(sign){
		case SIGPIPE:
		  have_sigpipe = 1;
		  break;
		case SIGHUP:
		  have_sighup = 1;
		  break;
	        case SIGCHLD:
		  wait(&status);
		  break;
		case SIGINT:
		  have_sigint=1;
		  break;
		case SIGTERM:
		  have_sigterm=1;
		  break;
		default:
	}
	init_signals();
}

ssize_t recvall(int s, void *buf, size_t len, int flags);

ssize_t recvall(int s, void *buf, size_t len, int flags) {
  size_t received = 0;
  char *b = buf;
  while ((received += recv(s, &b[received], len - received, flags)) < len);
  return received;
}

#define UDM_STOREIND_SIG "msINDEX\0"
#define UDM_STORE_SIG "msSTORE\0"
#define UDM_SIG_LEN 8
#define UDM_STORE_BITS 0xFFFF      /* bits of rec_id for file no. */
#define UDM_STORE_MASK ((1L << UDM_STORE_BITS) - 1)

#define UDM_HASH_PRIME 4093
#define UDM_HASH(x)    ((x) % UDM_HASH_PRIME)

typedef struct StoreItem {
  unsigned int rec_id;
  long offset;
  long next;
  size_t size;
} UDM_STOREITEM;


static struct flock* file_lock(int type, int whence) {
    static struct flock ret ;
    ret.l_type = type ;
    ret.l_start = 0 ;
    ret.l_whence = whence ;
    ret.l_len = 0 ;
    ret.l_pid = getpid() ;
    return &ret ;
}

static void read_lock(FILE *f) {   /* a shared lock on an entire file */

    fcntl(fileno(f), F_SETLKW, file_lock(F_RDLCK, SEEK_SET));
}

static void write_lock(FILE *f) {  /* an exclusive lock on an entire file */

    fcntl(fileno(f), F_SETLKW, file_lock(F_WRLCK, SEEK_SET));
}

static void un_lock(FILE *f) { /* ulock an entire file */
  fcntl(fileno(f), F_SETLKW, file_lock(F_UNLCK, SEEK_SET));
}


#endif

int main(int argc, char **argv) {

#ifdef HAVE_ZLIB

        fd_set mask;
	struct sockaddr_in server_addr, client_addr;
	const char *config_name= UDM_CONF_DIR "/stored.conf";
	const char *store_dir;
	int pid_fd, nport, s, on = 1;
	char pidbuf[1024];
	char buf[1024];
	int log2stderr=1;

	/* Check that another instance isn't running */
	/* and create PID file.                      */

	sprintf(pidname,"%s/%s",UDM_VAR_DIR,"stored.pid");
	pid_fd = open(pidname,O_CREAT|O_EXCL|O_WRONLY,0644);
	if(pid_fd < 0){
		fprintf(stderr,"%s Can't create '%s': %s\n", time_pid_info(), pidname, strerror(errno));
		if(errno == EEXIST){
			fprintf(stderr,"It seems that another stored is already running!\n");
			fprintf(stderr,"Remove '%s' if it is not true.\n",pidname);
 		}
		exit(1);
	}

	sprintf(pidbuf,"%d\n",(int)getpid());
	write(pid_fd,&pidbuf,strlen(pidbuf));

	atexit(&exitproc);
	init_signals();
	UdmInit(); /* Initialize library */

	Conf = UdmAllocEnv();

	UdmLoadConfig(Conf,config_name,0,0);

	UdmOpenLog("stored",Conf,log2stderr);

	if(UdmEnvErrCode(Conf)){
 		fprintf(stderr,"%s\n",UdmEnvErrMsg(Conf));
		UdmFreeEnv(Conf);
		unlink(pidname);
		exit(1);
	}
	Agent = UdmAllocAgent(Conf,0,0);

	nport = UDM_STORED_PORT;

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  UdmLog(Agent, verb, "%s socket() error %d\n",time_pid_info(),errno);
	  UdmFreeAgent(Agent);
	  UdmFreeEnv(Conf);
	  unlink(pidname);
	  exit(1);
	}
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) != 0){
		UdmLog(Agent, verb, "%s setsockopt() error %d\n",time_pid_info(),errno);
		UdmFreeAgent(Agent);
		UdmFreeEnv(Conf);
		unlink(pidname);
		exit(1);
	}

	bzero(&server_addr, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	if(Agent->Conf->listen){
	  char * cport;
			
	  if((cport=strchr(Agent->Conf->listen,':'))){
	    UdmLog(Agent,verb,"Listening '%s'",Agent->Conf->listen);
	    *cport='\0';
	    server_addr.sin_addr.s_addr = inet_addr(Agent->Conf->listen);
	    nport=atoi(cport+1);
	  }else{
	    nport=atoi(Agent->Conf->listen);
	    server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	    UdmLog(Agent,verb,"Listening port %d",nport);
	  }
	}else{
	  UdmLog(Agent,verb,"Listening port %d",nport);
	  server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	}
	server_addr.sin_port = htons((u_short)nport);
	

	if (bind(s, (struct sockaddr *)&server_addr, sizeof(server_addr)) == -1) {
		fprintf(stderr,"%s bind() error %d %s\n",time_pid_info(),errno,strerror(errno));
		unlink(pidname);
		UdmFreeAgent(Agent);
		UdmFreeEnv(Conf);
		exit(1);
	}

	if (listen(s, UDM_STORED_MAXCLNTS) == -1) {
		fprintf(stderr,"%s listen() error %d %s\n",time_pid_info(),errno,strerror(errno));
		unlink(pidname);
		UdmFreeAgent(Agent);
		UdmFreeEnv(Conf);
		exit(1);
	}
	
	FD_ZERO(&mask);
	FD_SET(s, &mask);

	while(1) {
	  size_t addrlen;
	  int ns, pid, sel;
	  struct timeval tval;
	  fd_set msk;
	  
		tval.tv_sec = 300;
		tval.tv_usec = 0;
		msk = mask;
		sel = select(16, &msk, 0, 0, &tval);

		if(have_sighup){
			UdmLog(Agent,verb,"SIGHUP arrived. Reloading config.");
			have_sighup=0;
			UdmFreeAgent(Agent);
			UdmFreeEnv(Conf);
			Conf = UdmAllocEnv();

			UdmLoadConfig(Conf,config_name,0,0);

			if(UdmEnvErrCode(Conf)) {
			  fprintf(stderr,"%s\n",UdmEnvErrMsg(Conf));
			  UdmFreeEnv(Conf);
			  unlink(pidname);
			  exit(1);
			}
			Agent = UdmAllocAgent(Conf,0,0);
		}
		if(have_sigint){
			UdmLog(Agent,verb,"SIGINT arrived");
			have_sigint=0;
			break;
		}
		if(have_sigterm){
			UdmLog(Agent,verb,"SIGTERM arrived");
			have_sigterm=0;
			break;
		}
		if(have_sigpipe){
			UdmLog(Agent,verb,"SIGPIPE arrived. Broken pipe !");
			have_sigpipe=0;
			break;
		}

		if(sel==0)continue;
		if(sel==-1){
			switch(errno){
				case EINTR:	/* Child */
					break;
			        default:
					UdmLog(Agent,verb,"FIXME select error %d %s",errno,strerror(errno));
			}
			continue;
		}


		bzero(&client_addr, addrlen = sizeof(client_addr));

	  if ((ns = accept(s, (struct sockaddr *) &client_addr, &addrlen)) == -1) {
	    fprintf(stderr,"%s accept() error %d %s\n",time_pid_info(),errno,strerror(errno));
	    unlink(pidname);
	    exit(1);
	  }

#define Client inet_ntoa(client_addr.sin_addr)

	  UdmLog(Agent, verb, "[%s] Accept", Client);
    
	  if ((pid = fork() ) == -1) {
	    fprintf(stderr,"%s fork() error %d %s\n",time_pid_info(),errno,strerror(errno));
	    unlink(pidname);
	    exit(1);
	  }

	  if (FD_ISSET(s, &msk)) {

	    if (pid == 0) { /* child process */
	      unsigned int rec_id, FileNo, hash, mishash = 0;
	      FILE  *Ifd = NULL, *Sfd = NULL;
	      char *Ifilename = NULL, *Sfilename = NULL;
	      size_t DocSize = 0, filenamelen;
	      ssize_t nbytes;
	      Byte *Doc = NULL, *CDoc = NULL;
	      long CurrentItemPos, NewItemPos;
	      z_stream zstream;
	      UDM_STOREITEM Item, *hTable;
	      char sig[8];
	    

	      close(s);
      
	      if ((nbytes = recv(ns, buf, 1, 0))) {
		if (strncmp(buf, "S", 1) == 0) { /* Store document */
		
		  if (recvall(ns, &rec_id, sizeof(rec_id), MSG_WAITALL) < 0) {
		    closesocket(ns);
		    exit(1);
		  }

		  if (recvall(ns, &DocSize, sizeof(DocSize), MSG_WAITALL) < 0) {
		    closesocket(ns);
		    exit(1);
		  }
		  Doc = (Byte *) UdmXmalloc(DocSize);
		  if (Doc == NULL) {
		    closesocket(ns);
		    exit(1);
		  }
		  if (recvall(ns, Doc, DocSize, MSG_WAITALL) < 0) {
		    closesocket(ns);
		    exit(1);
		  }

		  send (ns, "STORED", 6, 0); 
		  close(ns); 

		  /* communications end */

		  zstream.zalloc = Z_NULL;
		  zstream.zfree = Z_NULL;
		  zstream.opaque = Z_NULL;
		  zstream.next_in = Doc;
		
		  if (deflateInit2(&zstream, 9, Z_DEFLATED, 15, 9, Z_DEFAULT_STRATEGY) == Z_OK) {
		
		    zstream.avail_in = DocSize;
		    zstream.avail_out = 2 * DocSize;
		    CDoc = zstream.next_out = (Byte *) UdmXmalloc(2 * DocSize);
		    if (zstream.next_out == NULL) {
		      exit(1);
		    }
		    deflate(&zstream, Z_FINISH);
		    deflateEnd(&zstream);


/* store operations */

#define ABORT	    UDM_FREE(Doc); \
                    UDM_FREE(CDoc); \
		    UDM_FREE(Ifilename); \
		    UDM_FREE(Sfilename); \
                    un_lock(Sfd); \
                    un_lock(Ifd); \
                    fclose(Sfd); \
                    fclose(Ifd); \
		    exit (0);


		    FileNo =  (rec_id & UDM_STORE_BITS);
		    hash = UDM_HASH(rec_id);
		    filenamelen = ((Conf->vardir == NULL) ? strlen(UDM_VAR_DIR "/store/") : strlen(Conf->vardir) + 7) + 34;
		    store_dir = (Conf->vardir == NULL) ? UDM_VAR_DIR : Conf->vardir;
		    if (
			((Ifilename = (char *)UdmXmalloc(filenamelen)) == NULL) ||
			((Sfilename = (char *)UdmXmalloc(filenamelen)) == NULL) 
			) {
		      UDM_FREE(Doc);
		      UDM_FREE(CDoc);
		      exit(0);
		    }
		    sprintf(Sfilename, "%s/store/%04x.s", store_dir, FileNo);
		    sprintf(Ifilename, "%s/store/%04x.i", store_dir, FileNo);

		    if ((Ifd = fopen(Ifilename, "rb+")) == NULL) {
		      if ((Ifd = fopen(Ifilename, "wb+")) == NULL) {
			fprintf(stderr, "Can't open/create file %s", Ifilename);
			UDM_FREE(Doc); 
			UDM_FREE(CDoc);
			UDM_FREE(Ifilename); 
			UDM_FREE(Sfilename); 
			exit (0);
		      }
		      if ((hTable = (UDM_STOREITEM *)UdmXmalloc(sizeof(UDM_STOREITEM) * UDM_HASH_PRIME)) == NULL) {
			ABORT;
		      }
		      bzero(hTable, sizeof(UDM_STOREITEM) * UDM_HASH_PRIME);
		      if (
			  (fwrite(UDM_STOREIND_SIG, UDM_SIG_LEN, 1, Ifd) != 1) ||
			  (fwrite(hTable, sizeof(UDM_STOREITEM), UDM_HASH_PRIME, Ifd) != UDM_HASH_PRIME) 
			  ) {
			fprintf(stderr, "Can't set new index for file %s\n", Ifilename);
			UDM_FREE(hTable);
			ABORT;
		      }
		      UDM_FREE(hTable);
		      if (fseek(Ifd, 0, SEEK_SET)) {
			ABORT;
		      }
		    }
		    write_lock(Ifd);

		    if ((Sfd = fopen(Sfilename, "rb+")) == NULL) {
		      if ((Sfd = fopen(Sfilename, "wb+")) == NULL) {
			fprintf(stderr, "Can't open/create file %s", Sfilename);
			UDM_FREE(Doc); 
			UDM_FREE(CDoc);
			UDM_FREE(Ifilename); 
			UDM_FREE(Sfilename);
			un_lock(Ifd);
			fclose(Ifd);
			exit (0);
		      }
		      if (fwrite(UDM_STORE_SIG, UDM_SIG_LEN, 1, Sfd) != 1) { 
			fprintf(stderr, "Can't set signature for file %s\n", Sfilename);
			ABORT;
		      }
		      if (fseek(Sfd, 0, SEEK_SET)) {
			ABORT;
		      }
		    }
		    write_lock(Sfd);

		    /* check store signature or set for new */
		    if (fread(sig, UDM_SIG_LEN, 1, Sfd) != 1) {
		      fprintf(stderr, "Can't read signature for file %s\n", Sfilename);
		      ABORT;
		    }
		    if ( strcmp(sig, UDM_STORE_SIG) != 0 ) {
		      fprintf(stderr, "Can't check signature %s for file %s\n", sig, Sfilename);
		      ABORT;
		    }
		    /* check index signature or set for new */
		    if  (fread(sig, UDM_SIG_LEN, 1, Ifd) != 1) {/* can't read signature, try to write one */
		      fprintf(stderr, "Can't read signature for file %s\n", Ifilename);
		      ABORT;
		    }
		    if ( strcmp(sig, UDM_STOREIND_SIG) != 0 ) {
		      fprintf(stderr, "Can't check signature for file %s\n", Ifilename);
		      ABORT;
		    }

		    /* search rec_id */

		    if (fseek(Ifd, (long)(UDM_SIG_LEN + hash * sizeof(UDM_STOREITEM)), SEEK_SET)) {
		      ABORT;
		    }
		    CurrentItemPos = ftell(Ifd);
		    if (fread(&Item, sizeof(UDM_STOREITEM), 1, Ifd) != 1) {
		      fprintf(stderr, "Can't read index for file %s %ld\n", Ifilename, CurrentItemPos);
		      ABORT;
		    }

		    while((Item.next != 0) && (Item.rec_id != rec_id)) {
		      CurrentItemPos = Item.next;
		      if (fseek(Ifd, CurrentItemPos, SEEK_SET)) {
			ABORT;
		      }
		      if (fread(&Item, sizeof(UDM_STOREITEM), 1, Ifd) != 1) {
			fprintf(stderr, "Can't read hash chain for file %s\n", Ifilename);
			ABORT;
		      }
		      mishash = 1;
		    }

		    if (Item.rec_id == rec_id) {
		      if (Item.size < zstream.total_out) {
			if (fseek(Sfd, 0, SEEK_END)) {
			  ABORT;
			}
			Item.offset = ftell(Sfd);
		      } else {
			if (fseek(Sfd, Item.offset, SEEK_SET)) {
			  ABORT;
			}
		      }
		    } else { /* new rec_id added */
		      if (mishash) {
			if (fseek(Ifd, 0, SEEK_END)) {
			  ABORT;
			}
			Item.next = NewItemPos = ftell(Ifd);
			if (fseek(Ifd, CurrentItemPos, SEEK_SET)) {
			  ABORT;
			}
			if (fwrite(&Item, sizeof(UDM_STOREITEM), 1, Ifd) != 1) {
			  ABORT;
			}
			CurrentItemPos = NewItemPos;
		      }
		      Item.rec_id = rec_id;
		      Item.next = 0;
		      if (fseek(Sfd, 0, SEEK_END)) {
			ABORT;
		      }
		      Item.offset = ftell(Sfd);
		    }
		    Item.size = zstream.total_out;
		    if (fwrite(CDoc, zstream.total_out, 1,  Sfd) != 1) {
		      ABORT;
		    }
		    if (fseek(Ifd, CurrentItemPos, SEEK_SET)) {
		      ABORT;
		    }
		    if (fwrite(&Item, sizeof(UDM_STOREITEM), 1, Ifd) != 1) {
		      fprintf(stderr, "Can't write index for file %s\n", Ifilename);
		    }

		    fprintf(stderr, "[%s] Stored rec_id: %x Size: %d Ratio: %5.2f%%\n", Client,
			    rec_id, DocSize, 100.0 * zstream.total_out / DocSize);
		    ABORT;
/* /store operations */

		  }

		  ABORT;

		} else if(strncmp(buf, "G", 1) == 0) { /* retrieve document */
		  if (recvall(ns, &rec_id, sizeof(rec_id), MSG_WAITALL) < 0) {
		    closesocket(ns);
		    exit(1);
		  }

		  FileNo =  (rec_id & UDM_STORE_BITS);
		  hash = UDM_HASH(rec_id);
		  filenamelen = ((Conf->vardir == NULL) ? strlen(UDM_VAR_DIR "/store/") : strlen(Conf->vardir) + 7) + 34;
		  store_dir = (Conf->vardir == NULL) ? UDM_VAR_DIR : Conf->vardir;
		  if (
		      ((Ifilename = (char *)UdmXmalloc(filenamelen)) == NULL) ||
		      ((Sfilename = (char *)UdmXmalloc(filenamelen)) == NULL) 
		      ) {
		    UDM_FREE(Doc);
		    UDM_FREE(CDoc);
                    send(ns, &DocSize, sizeof(DocSize), 0); 
                    closesocket(ns); 
		    exit(0);
		  }
		  sprintf(Sfilename, "%s/store/%04x.s", store_dir, FileNo);
		  sprintf(Ifilename, "%s/store/%04x.i", store_dir, FileNo);

		  if ((Ifd = fopen(Ifilename, "rb+")) == NULL) {
		    if ((Ifd = fopen(Ifilename, "wb+")) == NULL) {
		      fprintf(stderr, "Can't open/create file %s", Ifilename);
		      UDM_FREE(Doc); 
		      UDM_FREE(CDoc);
		      UDM_FREE(Ifilename); 
		      UDM_FREE(Sfilename); 
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      exit (0);
		    }
		    if ((hTable = (UDM_STOREITEM *)UdmXmalloc(sizeof(UDM_STOREITEM) * UDM_HASH_PRIME)) == NULL) {
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		    bzero(hTable, sizeof(UDM_STOREITEM) * UDM_HASH_PRIME);
		    if (
			(fwrite(UDM_STOREIND_SIG, UDM_SIG_LEN, 1, Ifd) != 1) ||
			(fwrite(hTable, sizeof(UDM_STOREITEM), UDM_HASH_PRIME, Ifd) != UDM_HASH_PRIME) 
			) {
		      fprintf(stderr, "Can't set new index for file %s\n", Ifilename);
		      UDM_FREE(hTable);
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		    UDM_FREE(hTable);
		    if (fseek(Ifd, 0, SEEK_SET)) {
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		  }
		  read_lock(Ifd);

		  if ((Sfd = fopen(Sfilename, "rb+")) == NULL) {
		    if ((Sfd = fopen(Sfilename, "wb+")) == NULL) {
		      fprintf(stderr, "Can't open/create file %s", Sfilename);
		      UDM_FREE(Doc); 
		      UDM_FREE(CDoc);
		      UDM_FREE(Ifilename); 
		      UDM_FREE(Sfilename);
		      un_lock(Ifd);
		      fclose(Ifd);
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      exit (0);
		    }
		    if (fwrite(UDM_STORE_SIG, UDM_SIG_LEN, 1, Sfd) != 1) { 
		      fprintf(stderr, "Can't set signature for file %s\n", Sfilename);
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		    if (fseek(Sfd, 0, SEEK_SET)) {
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		  }
		  read_lock(Sfd);

		  /* check store signature or set for new */
		  if (fread(sig, UDM_SIG_LEN, 1, Sfd) != 1) {
		    fprintf(stderr, "Can't read signature for file %s\n", Sfilename);
		    send(ns, &DocSize, sizeof(DocSize), 0); 
		    closesocket(ns); 
		    ABORT;
		  }
		  if ( strcmp(sig, UDM_STORE_SIG) != 0 ) {
		    fprintf(stderr, "Can't check signature %s for file %s\n", sig, Sfilename);
		    send(ns, &DocSize, sizeof(DocSize), 0); 
		    closesocket(ns); 
		    ABORT;
		  }
		  /* check index signature or set for new */
		  if  (fread(sig, UDM_SIG_LEN, 1, Ifd) != 1) {/* can't read signature, try to write one */
		    fprintf(stderr, "Can't read signature for file %s\n", Ifilename);
		    send(ns, &DocSize, sizeof(DocSize), 0); 
		    closesocket(ns); 
		    ABORT;
		  }
		  if ( strcmp(sig, UDM_STOREIND_SIG) != 0 ) {
		    fprintf(stderr, "Can't check signature for file %s\n", Ifilename);
		    send(ns, &DocSize, sizeof(DocSize), 0); 
		    closesocket(ns); 
		    ABORT;
		  }

		  /* search rec_id */

		  if (fseek(Ifd, (long)(UDM_SIG_LEN + hash * sizeof(UDM_STOREITEM)), SEEK_SET)) {
		    send(ns, &DocSize, sizeof(DocSize), 0); 
		    closesocket(ns); 
		    ABORT;
		  }
		  CurrentItemPos = ftell(Ifd);
		  if (fread(&Item, sizeof(UDM_STOREITEM), 1, Ifd) != 1) {
		    fprintf(stderr, "Can't read index for file %s at unstore\n", Ifilename);
		    send(ns, &DocSize, sizeof(DocSize), 0); 
		    closesocket(ns); 
		    ABORT;
		  }

		  while((Item.next != 0) && (Item.rec_id != rec_id)) {
		    CurrentItemPos = Item.next;
		    if (fseek(Ifd, CurrentItemPos, SEEK_SET)) {
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		    if (fread(&Item, sizeof(UDM_STOREITEM), 1, Ifd) != 1) {
		      fprintf(stderr, "Can't read hash chain for file %s\n", Ifilename);
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		    mishash = 1;
		  }
		  
		  if (Item.rec_id == rec_id) {
		    if (fseek(Sfd, Item.offset, SEEK_SET)) {
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		    zstream.avail_in = DocSize = Item.size;
		    zstream.avail_out = UDM_MAXDOCSIZE;
		    CDoc = zstream.next_in = (Byte *) UdmXmalloc(DocSize);
		    Doc = zstream.next_out = (Byte *) UdmXmalloc(UDM_MAXDOCSIZE);
		    if (CDoc == NULL || Doc == NULL) {
		      DocSize = 0;
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		    zstream.zalloc = Z_NULL;
		    zstream.zfree = Z_NULL;
		    zstream.opaque = Z_NULL;
		    if ((fread(CDoc, DocSize, 1, Sfd) != 1)
			|| (inflateInit2(&zstream, 15) != Z_OK)) {
		      DocSize = 0;
		      send(ns, &DocSize, sizeof(DocSize), 0); 
		      closesocket(ns); 
		      ABORT;
		    }
		    inflate(&zstream, Z_FINISH);
		    inflateEnd(&zstream);
		    DocSize = zstream.total_out;
		    send(ns, &DocSize, sizeof(DocSize), 0);
		    send(ns, Doc, DocSize, 0);
		    closesocket(ns); 
		    
		    
		  } else {
		    send(ns, &DocSize, sizeof(DocSize), 0); 
		    closesocket(ns); 
		    fprintf(stderr, "[%s] Not found rec_id: %x\n", Client, rec_id);
		    ABORT;
		  }

		  fprintf(stderr, "[%s] Retrived rec_id: %x Size: %d Ratio: %5.2f%%\n", Client,
			  rec_id, DocSize, 100.0 * zstream.total_in / DocSize);
		  ABORT;



		} else {
		  fprintf(stderr,"%s <hello> error %d %s\n",time_pid_info(),errno,buf /*strerror(errno)*/);
		  close(ns);
		  exit(1);
		}
	      } else {
		fprintf(stderr,"%s recv() <hello> error %d %s\n",time_pid_info(),errno,strerror(errno));
		close(ns);
		exit(1);
	      }


	      close(ns);
	      exit(0);
	    }
	  }
	  /* parent process */
	  close(ns);
	}

	unlink(pidname);
	UdmFreeAgent(Agent);
	UdmFreeEnv(Conf);

#else
	fprintf(stderr, "zlib support required. Please rebuild with --with-zlib option\n");

#endif
	return 0;
}
